"""
TRedis
======
An asynchronous Redis client for Tornado

"""
from tredis.client import Client, RedisClient
from tredis.exceptions import *
from tredis.strings import BITOP_AND, BITOP_OR, BITOP_XOR, BITOP_NOT

__version__ = '0.8.0'



"""Redis Sorted Set Commands Mixin"""


class SortedSetsMixin(object):
    """Redis Sorted Set Commands Mixin"""

    def zadd(self, key, *members, **kwargs):
        """Adds all the specified members with the specified scores to the
        sorted set stored at key. It is possible to specify multiple score /
        member pairs. If a specified member is already a member of the sorted
        set, the score is updated and the element reinserted at the right
        position to ensure the correct ordering.

        If key does not exist, a new sorted set with the specified members as
        sole members is created, like if the sorted set was empty. If the key
        exists but does not hold a sorted set, an error is returned.

        The score values should be the string representation of a double
        precision floating point number. +inf and -inf values are valid values
        as well.

        **Members parameters**

        ``members`` could be either:
        - a single dict where keys correspond to scores and values to elements
        - multiple strings paired as score then element

        .. code:: python

            yield client.zadd('myzset', {'1': 'one', '2': 'two'})
            yield client.zadd('myzset', '1', 'one', '2', 'two')

        **ZADD options (Redis 3.0.2 or greater)**

        ZADD supports a list of options. Options are:

        - ``xx``: Only update elements that already exist. Never add elements.
        - ``nx``: Don't update already existing elements. Always add new
            elements.
        - ``ch``: Modify the return value from the number of new elements
            added, to the total number of elements changed (CH is an
            abbreviation of changed). Changed elements are new elements added
            and elements already existing for which the score was updated. So
            elements specified in the command having the same score as they had
            in the past are not counted. Note: normally the return value of
            ``ZADD`` only counts the number of new elements added.
        - ``incr``: When this option is specified ``ZADD`` acts like
            :meth:`~tredis.RedisClient.zincrby`. Only one score-element pair
            can be specified in this mode.

        .. note::

           **Time complexity**: ``O(log(N))`` for each item added, where ``N``
           is the number of elements in the sorted set.

        :param key: The key of the sorted set
        :type key: :class:`str`, :class:`bytes`
        :param members: Elements to add
        :type members: :class:`dict`, :class:`str`, :class:`bytes`
        :keyword bool xx: Only update elements that already exist
        :keyword bool nx: Don't update already existing elements
        :keyword bool ch: Return the number of changed elements
        :keyword bool incr: Increment the score of an element
        :rtype: int, :class:`str`, :class:`bytes`
        :returns: Number of elements changed, or the new score if incr is set
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        xx = kwargs.pop('xx', False)
        nx = kwargs.pop('nx', False)
        ch = kwargs.pop('ch', False)
        incr = kwargs.pop('incr', False)
        command = [b'ZADD', key]
        if xx:
            command += ['XX']
        if nx:
            command += ['NX']
        if ch:
            command += ['CH']
        if incr:
            command += ['INCR']

        if len(members) == 1:
            for k in members[0]:
                command += [k, members[0][k]]
        else:
            command += list(members)
        return self._execute(command)

    def zcard(self, key):
        """Returns the set cardinality (number of elements) of the sorted set
        stored at key.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key of the set
        :type key: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'ZCARD', key])

    def zrange(self, key, start=0, stop=-1, with_scores=False):
        """Returns the specified range of elements in the sorted set stored at
        key. The elements are considered to be ordered from the lowest to the
        highest score. Lexicographical order is used for elements with equal
        score.

        See :meth:`tredis.Client.zrevrange` when you need the elements ordered
        from highest to lowest score (and descending lexicographical order for
        elements with equal score).

        Both start and stop are zero-based indexes, where ``0`` is the first
        element, ``1`` is the next element and so on. They can also be negative
        numbers indicating offsets from the end of the sorted set, with ``-1``
        being the last element of the sorted set, ``-2`` the penultimate
        element and so on.

        ``start`` and ``stop`` are inclusive ranges, so for example
        ``ZRANGE myzset 0 1`` will return both the first and the second element
        of the sorted set.

        Out of range indexes will not produce an error. If start is larger than
        the largest index in the sorted set, or ``start > stop``, an empty list
        is returned. If stop is larger than the end of the sorted set Redis
        will treat it like it is the last element of the sorted set.

        It is possible to pass the ``WITHSCORES`` option in order to return the
        scores of the elements together with the elements. The returned list
        will contain ``value1,score1,...,valueN,scoreN`` instead of
        ``value1,...,valueN``. Client libraries are free to return a more
        appropriate data type (suggestion: an array with (value, score)
        arrays/tuples).

        .. note::

           **Time complexity**: ``O(log(N)+M)`` with ``N`` being the number of
           elements in the sorted set and ``M`` the number of elements
           returned.

        :param key: The key of the sorted set
        :type key: :class:`str`, :class:`bytes`
        :param int start: The starting index of the sorted set
        :param int stop: The ending index of the sorted set
        :param bool with_scores: Return the scores with the elements

        :rtype: list
        :raises: :exc:`~tredis.exceptions.RedisError`
        """
        command = [b'ZRANGE', key, start, stop]
        if with_scores:
            command += ['WITHSCORES']
        return self._execute(command)

    def zrangebyscore(self,
                      key,
                      min_score,
                      max_score,
                      with_scores=False,
                      offset=0,
                      count=0):
        """Returns all the elements in the sorted set at key with a score
        between min and max (including elements with score equal to min or
        max). The elements are considered to be ordered from low to high
        scores.

        The elements having the same score are returned in lexicographical
        order (this follows from a property of the sorted set implementation in
        Redis and does not involve further computation).

        The optional ``offset`` and ``count`` arguments can be used to only get
        a range of the matching elements (similar to SELECT LIMIT offset, count
        in SQL). Keep in mind that if offset is large, the sorted set needs to
        be traversed for offset elements before getting to the elements to
        return, which can add up to ``O(N)`` time complexity.

        The optional ``with_scores`` argument makes the command return both the
        element and its score, instead of the element alone. This option is
        available since Redis 2.0.

        **Exclusive intervals and infinity**

        ``min_score`` and ``max_score`` can be ``-inf`` and ``+inf``, so that
        you are not required to know the highest or lowest score in the sorted
        set to get all elements from or up to a certain score.

        By default, the interval specified by ``min_score`` and ``max_score``
        is closed (inclusive). It is possible to specify an open interval
        (exclusive) by prefixing the score with the character ``(``. For
        example:

        .. code::

            ZRANGEBYSCORE zset (1 5

        Will return all elements with ``1 < score <= 5`` while:

        .. code::

            ZRANGEBYSCORE zset (5 (10

        Will return all the elements with ``5 < score < 10`` (5 and 10
        excluded).

        .. note::

           **Time complexity**: ``O(log(N)+M)`` with ``N`` being the number of
           elements in the sorted set and ``M`` the number of elements being
           returned. If ``M`` is constant (e.g. always asking for the first
           10 elements with ``count``), you can consider it ``O(log(N))``.

        :param key: The key of the sorted set
        :type key: :class:`str`, :class:`bytes`
        :param min_score: Lowest score definition
        :type min_score: :class:`str`, :class:`bytes`
        :param max_score: Highest score definition
        :type max_score: :class:`str`, :class:`bytes`
        :param bool with_scores: Return elements and scores
        :param offset: The number of elements to skip
        :type min_score: :class:`str`, :class:`bytes`
        :param count: The number of elements to return
        :type min_score: :class:`str`, :class:`bytes`
        :rtype: list
        :raises: :exc:`~tredis.exceptions.RedisError`
        """
        command = [b'ZRANGEBYSCORE', key, min_score, max_score]
        if with_scores:
            command += ['WITHSCORES']
        if offset or count:
            command += ['LIMIT', offset, count]
        return self._execute(command)

    def zrem(self, key, *members):
        """Removes the specified members from the sorted set stored at key.
         Non existing members are ignored.

        An error is returned when key exists and does not hold a sorted set.

        .. note::

           **Time complexity**: ``O(M*log(N))`` with ``N`` being the number of
           elements in the sorted set and ``M`` the number of elements to be
           removed.

        :param key: The key of the sorted set
        :type key: :class:`str`, :class:`bytes`
        :param members: One or more member values to remove
        :type members: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`
        """
        return self._execute([b'ZREM', key] + list(members))

    def zremrangebyscore(self, key, min_score, max_score):
        """Removes all elements in the sorted set stored at key with a score
        between min and max.

        Intervals are described in :meth:`~tredis.RedisClient.zrangebyscore`.

        Returns the number of elements removed.

        .. note::

           **Time complexity**: ``O(log(N)+M)`` with ``N`` being the number of
           elements in the sorted set and M the number of elements removed by
           the operation.

        :param key: The key of the sorted set
        :type key: :class:`str`, :class:`bytes`
        :param min_score: Lowest score definition
        :type min_score: :class:`str`, :class:`bytes`
        :param max_score: Highest score definition
        :type max_score: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`
        """
        return self._execute([b'ZREMRANGEBYSCORE', key, min_score, max_score])

    def zrevrange(self, key, start=0, stop=-1, with_scores=False):
        """Returns the specified range of elements in the sorted set stored at
        key. The elements are considered to be ordered from the highest to the
        lowest score. Descending lexicographical order is used for elements
        with equal score.

        Apart from the reversed ordering, :py:meth:`~tredis.Client.zrevrange`
        is similar to :py:meth:`~tredis.Client.zrange` .

        .. note::

           **Time complexity**: ``O(log(N)+M)`` with ``N`` being the number of
           elements in the sorted set and ``M`` the number of elements
           returned.

        :param key: The key of the sorted set
        :type key: :class:`str`, :class:`bytes`
        :param int start: The starting index of the sorted set
        :param int stop: The ending index of the sorted set
        :param bool with_scores: Return the scores with the elements

        :rtype: list
        :raises: :exc:`~tredis.exceptions.RedisError`
        """
        command = [b'ZREVRANGE', key, start, stop]
        if with_scores:
            command += ['WITHSCORES']
        return self._execute(command)

    def zscore(self, key, member):
        """Returns the score of member in the sorted set at key.
        If member does not exist in the sorted set, or key does not exist
        None is returned.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key of the set to check for membership in
        :type key: :class:`str`, :class:`bytes`
        :param member: The value to check for set membership with
        :type member: :class:`str`, :class:`bytes`
        :rtype: str or None
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'ZSCORE', key, member])



"""TRedis Exceptions"""


class TRedisException(Exception):
    """Raised as a top-level exception class for all exceptions raised by
    :class:`~tredis.RedisClient`.

    """
    pass


class ConnectError(TRedisException):
    """Raised when :class:`~tredis.RedisClient` can not connect to the
    specified Redis server.

    """
    pass


class ConnectionError(TRedisException):
    """Raised when :class:`~tredis.RedisClient` has had its connection to the
    Redis server interrupted unexpectedly.

    """
    pass


class AuthError(TRedisException):
    """Raised when :meth:`~tredis.RedisClient.auth` is invoked and the Redis
    server returns an error.

    """
    pass


class RedisError(TRedisException):
    """Raised when the Redis server returns a error to
    :class:`~tredis.RedisClient`. The string representation of this class will
    contain the error response from the Redis server, if one is sent.

    """
    pass


class SubscribedError(TRedisException):
    """Raised when a client is subscribed via
    :meth:`~tredis.RedisClient.subscribe` or
    :meth:`~tredis.RedisClient.psubscribe` and a command other than
    :meth:`~tredis.RedisClient.subscribe`,
    :meth:`~tredis.RedisClient.unsubscribe`,
    :meth:`~tredis.RedisClient.psubscribe`, or
    :meth:`~tredis.RedisClient.punsubscribe` was requested. Once the client
    enters the subscribed state it is not supposed to issue any other commands.

    """
    pass


class InvalidClusterCommand(TRedisException):
    """Raised when a method is invoked that is not able to be used when
    acting as a client for a Redis cluster.

    """
    pass



"""Redis Cluster Commands Mixin"""
import collections

from tredis import common

ClusterNode = collections.namedtuple('ClusterNode', [
    'id', 'ip', 'port', 'flags', 'master', 'ping_sent', 'pong_recv',
    'config_epoch', 'link_state', 'slots'
])
""":class:`tredis.cluster.ClusterNode` is a :class:`~collections.namedtuple`
that contains the attributes for a single node returned by the
``CLUSTER NODES`` command.

.. versionadded: 0.7

:param bytes id: The node ID
:param bytes ip: The IP address of the node
:param int port: The node TCP port
:param bytes flags: A list of comma separated flags: ``myself``, ``master``,
    ``slave``, ``fail?``, ``fail``, ``handshake``, ``noaddr``, ``noflags``.
:param bytes master: If the node is a slave, and the master is known, the master
    node ID, otherwise the ``-`` character.
:param int ping_sent: Milliseconds unix time at which the currently active ping
    was sent, or zero if there are no pending pings.
:param int pong_recv: Milliseconds unix time the last pong was received.
:param int config_epoch: The configuration epoch (or version) of the current
    node (or of the current master if the node is a slave). Each time there is
    a failover, a new, unique, monotonically increasing configuration epoch is
    created. If multiple nodes claim to serve the same hash slots, the one with
    higher configuration epoch wins.
:param bytes link_state: The state of the link used for the node-to-node cluster
    bus. We use this link to communicate with the node. Can be ``connected`` or
    ``disconnected``.
:param slots: A hash slot number or range. There may be up to 16384 entries in
    total (limit never reached). This is the list of hash slots served by this
    node. If the entry is just a number, is parsed as such. If it is a range,
    it is in the form start-end, and means that the node is responsible for
    all the hash slots from start to end including the start and end values.
:type slots: list(tuple(int, int))

"""


class ClusterMixin(object):
    """Redis Cluster Commands Mixin"""

    def cluster_add_slots(self, *slots):
        pass

    def cluster_count_failure_report(self, node):
        pass

    def cluster_count_keys_in_slot(self, slot):
        pass

    def cluster_del_slots(self, slot):
        pass

    def cluster_failover(self, method):
        pass  # force, takeover

    def cluster_forget(self, node_id):
        pass

    def cluster_get_keys_in_slot(self, slot, count):
        pass

    def cluster_info(self):
        """``CLUSTER INFO`` provides ``INFO`` style information about Redis
        Cluster vital parameters.

        .. versionadded:: 0.7.0

        :returns: A dictionary of current cluster information
        :rtype: dict

        :key cluster_state: State is ok if the node is able to receive
            queries. fail if there is at least one hash slot which is unbound
            (no node associated), in error state (node serving it is flagged
            with ``FAIL`` flag), or if the majority of masters can't be
            reached by this node.
        :key cluster_slots_assigned: Number of slots which are associated to
            some node (not unbound). This number should be ``16384`` for the
            node to work properly, which means that each hash slot should be
            mapped to a node.
        :key cluster_slots_ok: Number of hash slots mapping to a node not in
            ``FAIL`` or ``PFAIL`` state.
        :key cluster_slots_pfail: Number of hash slots mapping to a node in
            ``PFAIL`` state. Note that those hash slots still work
            correctly, as long as the ``PFAIL`` state is not promoted to
            ``FAIL`` by the failure detection algorithm. ``PFAIL``
            only means that we are currently not able to talk with the node,
            but may be just a transient error.
        :key cluster_slots_fail: Number of hash slots mapping to a node in
            ``FAIL`` state. If this number is not zero the node is not
            able to serve queries unless cluster-require-full-coverage is set
            to no in the configuration.
        :key cluster_known_nodes: The total number of known nodes in the
            cluster, including nodes in ``HANDSHAKE`` state that may not
            currently be proper members of the cluster.
        :key cluster_size: The number of master nodes serving at least one
            hash slot in the cluster.
        :key cluster_current_epoch: The local Current Epoch variable. This is
            used in order to create unique increasing version numbers during
            fail overs.
        :key cluster_my_epoch: The Config Epoch of the node we are talking
            with. This is the current configuration version assigned to this
            node.
        :key cluster_stats_messages_sent: Number of messages sent via the
            cluster node-to-node binary bus.
        :key cluster_stats_messages_received: Number of messages received via
            the cluster node-to-node binary bus.
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute(
            [b'CLUSTER', 'INFO'], format_callback=common.format_info_response)

    def cluster_key_slot(self, key):
        pass

    def cluster_meet(self, ip, port):
        pass

    def cluster_nodes(self):
        """Each node in a Redis Cluster has its view of the current cluster
        configuration, given by the set of known nodes, the state of the
        connection we have with such nodes, their flags, properties and
        assigned slots, and so forth.

        ``CLUSTER NODES`` provides all this information, that is, the current
        cluster configuration of the node we are contacting, in a serialization
        format which happens to be exactly the same as the one used by Redis
        Cluster itself in order to store on disk the cluster state (however the
        on disk cluster state has a few additional info appended at the end).

        Note that normally clients willing to fetch the map between Cluster
        hash slots and node addresses should use ``CLUSTER SLOTS`` instead.
        ``CLUSTER NODES``, that provides more information, should be used for
        administrative tasks, debugging, and configuration inspections. It is
        also used by ``redis-trib`` in order to manage a cluster.

        .. versionadded:: 0.7.0

        :rtype: list(:class:`~tredis.cluster.ClusterNode`)
        :raises: :exc:`~tredis.exceptions.RedisError`

        """

        def format_response(result):
            values = []
            for row in result.decode('utf-8').split('\n'):
                if not row:
                    continue
                parts = row.split(' ')
                slots = []
                for slot in parts[8:]:
                    if '-' in slot:
                        sparts = slot.split('-')
                        slots.append((int(sparts[0]), int(sparts[1])))
                    else:
                        slots.append((int(slot), int(slot)))
                ip_port = common.split_connection_host_port(parts[1])
                values.append(
                    ClusterNode(parts[0], ip_port[0], ip_port[1], parts[2],
                                parts[3], int(parts[4]), int(parts[5]),
                                int(parts[6]), parts[7], slots))
            return values

        return self._execute(
            ['CLUSTER', 'NODES'], format_callback=format_response)

    def cluster_replicate(self, node_id):
        pass

    def cluster_reset(self, method):
        pass  # hard, soft

    def cluster_save_config(self):
        pass

    def cluster_set_config_epoch(self, config_epoch):
        pass

    def cluster_set_slot(self, subcommand, node_id):
        pass

    def cluster_slaves(self, node_id):
        pass

    def cluster_slots(self):
        pass

    def cluster_readonly(self):
        pass

    def cluster_readwrite(self):
        pass



"""
Common utility methods

"""
import logging

LOGGER = logging.getLogger(__name__)


def maybe_raise_exception(future):
    if future.exception():
        raise future.exception()


def split_connection_host_port(value):
    parts = value.split(':')
    LOGGER.debug('Returning %r', (parts[0], int(parts[1])))
    return parts[0], int(parts[1])


def parse_info_value(value):
    """

    :param value:
    :return:

    """
    try:
        if '.' in value:
            return float(value)
        else:
            return int(value)
    except ValueError:
        if ',' in value or '=' in value:
            retval = {}
            for row in value.split(','):
                key, val = row.rsplit('=', 1)
                retval[key] = parse_info_value(val)
            return retval
        return value


def format_info_response(value):
    """Format the response from redis

    :param str value: The return response from redis
    :rtype: dict

    """
    info = {}
    for line in value.decode('utf-8').splitlines():
        if not line or line[0] == '#':
            continue
        if ':' in line:
            key, value = line.split(':', 1)
            info[key] = parse_info_value(value)
    return info



"""Redis Key Commands Mixin"""
from tornado import concurrent

# Python 2 support for ascii()
if 'ascii' not in dir(__builtins__):  # pragma: nocover
    from tredis.compat import ascii


class KeysMixin(object):
    """Redis Key Commands Mixin"""

    def delete(self, *keys):
        """Removes the specified keys. A key is ignored if it does not exist.
        Returns :data:`True` if all keys are removed.

        .. note::

           **Time complexity**: ``O(N)`` where ``N`` is the number of keys that
           will be removed. When a key to remove holds a value other than a
           string, the individual complexity for this key is ``O(M)`` where
           ``M`` is the number of elements in the list, set, sorted set or
           hash. Removing a single key that holds a string value is ``O(1)``.

        :param keys: One or more keys to remove
        :type keys: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'DEL'] + list(keys), len(keys))

    def dump(self, key):
        """Serialize the value stored at key in a Redis-specific format and
        return it to the user. The returned value can be synthesized back into
        a Redis key using the :meth:`~tredis.RedisClient.restore` command.

        The serialization format is opaque and non-standard, however it has a
        few semantic characteristics:

          - It contains a 64-bit checksum that is used to make sure errors
            will be detected. The :meth:`~tredis.RedisClient.restore` command
            makes sure to check the checksum before synthesizing a key using
            the serialized value.
          - Values are encoded in the same format used by RDB.
          - An RDB version is encoded inside the serialized value, so that
            different Redis versions with incompatible RDB formats will
            refuse to process the serialized value.
          - The serialized value does NOT contain expire information. In
            order to capture the time to live of the current value the
            :meth:`~tredis.RedisClient.pttl` command should be used.

        If key does not exist :data:`None` is returned.

        .. note::

           **Time complexity**: ``O(1)`` to access the key and additional
           ``O(N*M)`` to serialized it, where N is the number of Redis objects
           composing the value and ``M`` their average size. For small string
           values the time complexity is thus ``O(1)+O(1*M)`` where ``M`` is
           small, so simply ``O(1)``.

        :param key: The key to dump
        :type key: :class:`str`, :class:`bytes`
        :rtype: bytes, None

        """
        return self._execute([b'DUMP', key])

    def exists(self, key):
        """Returns :data:`True` if the key exists.

        .. note::

           **Time complexity**: ``O(1)``

        **Command Type**: String

        :param key: One or more keys to check for
        :type key: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'EXISTS', key], 1)

    def expire(self, key, timeout):
        """Set a timeout on key. After the timeout has expired, the key will
        automatically be deleted. A key with an associated timeout is often
        said to be volatile in Redis terminology.

        The timeout is cleared only when the key is removed using the
        :meth:`~tredis.RedisClient.delete` method or overwritten using the
        :meth:`~tredis.RedisClient.set` or :meth:`~tredis.RedisClient.getset`
        methods. This means that all the operations that conceptually alter the
        value stored at the key without replacing it with a new one will leave
        the timeout untouched. For instance, incrementing the value of a key
        with :meth:`~tredis.RedisClient.incr`, pushing a new value into a
        list with :meth:`~tredis.RedisClient.lpush`, or altering the field
        value of a hash with :meth:`~tredis.RedisClient.hset` are all
        operations that will leave the timeout untouched.

        The timeout can also be cleared, turning the key back into a persistent
        key, using the :meth:`~tredis.RedisClient.persist` method.

        If a key is renamed with :meth:`~tredis.RedisClient.rename`,
        the associated time to live is transferred to the new key name.

        If a key is overwritten by :meth:`~tredis.RedisClient.rename`, like in
        the case of an existing key ``Key_A`` that is overwritten by a call
        like ``client.rename(Key_B, Key_A)`` it does not matter if the original
        ``Key_A`` had a timeout associated or not, the new key ``Key_A`` will
        inherit all the characteristics of ``Key_B``.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to set an expiration for
        :type key: :class:`str`, :class:`bytes`
        :param int timeout: The number of seconds to set the timeout to
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute(
            [b'EXPIRE', key, ascii(timeout).encode('ascii')], 1)

    def expireat(self, key, timestamp):
        """:meth:`~tredis.RedisClient.expireat` has the same effect and
        semantic as :meth:`~tredis.RedisClient.expire`, but instead of
        specifying the number of seconds representing the TTL (time to live),
        it takes an absolute Unix timestamp (seconds since January 1, 1970).

        Please for the specific semantics of the command refer to the
        documentation of :meth:`~tredis.RedisClient.expire`.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to set an expiration for
        :type key: :class:`str`, :class:`bytes`
        :param int timestamp: The UNIX epoch value for the expiration
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute(
            [b'EXPIREAT', key,
             ascii(timestamp).encode('ascii')], 1)

    def keys(self, pattern):
        """Returns all keys matching pattern.

        While the time complexity for this operation is ``O(N)``, the constant
        times are fairly low. For example, Redis running on an entry level
        laptop can scan a 1 million key database in 40 milliseconds.

        .. warning:: Consider :meth:`~tredis.RedisClient.keys` as a
           command that should only be used in production environments with
           extreme care. It may ruin performance when it is executed against
           large databases. This command is intended for debugging and special
           operations, such as changing your keyspace layout. Don't use
           :meth:`~tredis.RedisClient.keys` in your regular application code.
           If you're looking for a way to find keys in a subset of your
           keyspace, consider using :meth:`~tredis.RedisClient.scan` or sets.

        Supported glob-style patterns:

         - ``h?llo`` matches ``hello``, ``hallo`` and ``hxllo``
         - ``h*llo`` matches ``hllo`` and ``heeeello``
         - ``h[ae]llo`` matches ``hello`` and ``hallo``, but not ``hillo``
         - ``h[^e]llo`` matches ``hallo``, ``hbllo``, but not ``hello``
         - ``h[a-b]llo`` matches ``hallo`` and ``hbllo``

        Use a backslash (``\``) to escape special characters if you want to
        match them verbatim.

        .. note::

           **Time complexity**: ``O(N)``

        :param pattern: The pattern to use when looking for keys
        :type pattern: :class:`str`, :class:`bytes`
        :rtype: list
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'KEYS', pattern])

    def migrate(self,
                host,
                port,
                key,
                destination_db,
                timeout,
                copy=False,
                replace=False):
        """Atomically transfer a key from a source Redis instance to a
        destination Redis instance. On success the key is deleted from the
        original instance and is guaranteed to exist in the target instance.

        The command is atomic and blocks the two instances for the time
        required to transfer the key, at any given time the key will appear to
        exist in a given instance or in the other instance, unless a timeout
        error occurs.

        .. note::

           **Time complexity**: This command actually executes a DUMP+DEL in
           the source instance, and a RESTORE in the target instance. See the
           pages of these commands for time complexity. Also an ``O(N)`` data
           transfer between the two instances is performed.

        :param host: The host to migrate the key to
        :type host: bytes, str
        :param int port: The port to connect on
        :param key: The key to migrate
        :type key: bytes, str
        :param int destination_db: The database number to select
        :param int timeout: The maximum idle time in milliseconds
        :param bool copy: Do not remove the key from the local instance
        :param bool replace: Replace existing key on the remote instance
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        command = [
            b'MIGRATE', host,
            ascii(port).encode('ascii'), key,
            ascii(destination_db).encode('ascii'),
            ascii(timeout).encode('ascii')
        ]
        if copy is True:
            command.append(b'COPY')
        if replace is True:
            command.append(b'REPLACE')
        return self._execute(command, b'OK')

    def move(self, key, db):
        """Move key from the currently selected database (see
        :meth:`~tredis.RedisClient.select`) to the specified destination
        database. When key already exists in the destination database, or it
        does not exist in the source database, it does nothing. It is possible
        to use :meth:`~tredis.RedisClient.move` as a locking primitive because
        of this.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to move
        :type key: :class:`str`, :class:`bytes`
        :param int db: The database number
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'MOVE', key, ascii(db).encode('ascii')], 1)

    def object_encoding(self, key):
        """Return the kind of internal representation used in order to store
        the value associated with a key

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to get the encoding for
        :type key: :class:`str`, :class:`bytes`
        :rtype: bytes
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'OBJECT', b'ENCODING', key])

    def object_idle_time(self, key):
        """Return the number of seconds since the object stored at the
        specified key is idle (not requested by read or write operations).
        While the value is returned in seconds the actual resolution of this
        timer is 10 seconds, but may vary in future implementations of Redis.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to get the idle time for
        :type key: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'OBJECT', b'IDLETIME', key])

    def object_refcount(self, key):
        """Return the number of references of the value associated with the
        specified key. This command is mainly useful for debugging.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to get the refcount for
        :type key: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'OBJECT', b'REFCOUNT', key])

    def persist(self, key):
        """Remove the existing timeout on key, turning the key from volatile
        (a key with an expire set) to persistent (a key that will never expire
        as no timeout is associated).

         .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to move
        :type key: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'PERSIST', key], 1)

    def pexpire(self, key, timeout):
        """This command works exactly like :meth:`~tredis.RedisClient.pexpire`
        but the time to live of the key is specified in milliseconds instead of
        seconds.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to set an expiration for
        :type key: :class:`str`, :class:`bytes`
        :param int timeout: The number of milliseconds to set the timeout to
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute(
            [b'PEXPIRE', key, ascii(timeout).encode('ascii')], 1)

    def pexpireat(self, key, timestamp):
        """:meth:`~tredis.RedisClient.pexpireat` has the same effect and
        semantic as :meth:`~tredis.RedisClient.expireat`, but the Unix time
        at which the key will expire is specified in milliseconds instead of
        seconds.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to set an expiration for
        :type key: :class:`str`, :class:`bytes`
        :param int timestamp: The expiration UNIX epoch value in milliseconds
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute(
            [b'PEXPIREAT', key,
             ascii(timestamp).encode('ascii')], 1)

    def pttl(self, key):
        """Like :meth:`~tredis.RedisClient.ttl` this command returns the
        remaining time to live of a key that has an expire set, with the sole
        difference that :meth:`~tredis.RedisClient.ttl` returns the amount of
        remaining time in seconds while :meth:`~tredis.RedisClient.pttl`
        returns it in milliseconds.

        In Redis 2.6 or older the command returns ``-1`` if the key does not
        exist or if the key exist but has no associated expire.

        Starting with Redis 2.8 the return value in case of error changed:

         - The command returns ``-2`` if the key does not exist.
         - The command returns ``-1`` if the key exists but has no associated
           expire.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to get the PTTL for
        :type key: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'PTTL', key])

    def randomkey(self):
        """Return a random key from the currently selected database.

        .. note::

           **Time complexity**: ``O(1)``

        :rtype: bytes
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'RANDOMKEY'])

    def rename(self, key, new_key):
        """Renames ``key`` to ``new_key``. It returns an error when the source
        and destination names are the same, or when ``key`` does not exist.
        If ``new_key`` already exists it is overwritten, when this happens
        :meth:`~tredis.RedisClient.rename` executes an implicit
        :meth:`~tredis.RedisClient.delete` operation, so if the deleted key
        contains a very big value it may cause high latency even if
        :meth:`~tredis.RedisClient.rename` itself is usually a constant-time
        operation.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to rename
        :type key: :class:`str`, :class:`bytes`
        :param new_key: The key to rename it to
        :type new_key: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'RENAME', key, new_key], b'OK')

    def renamenx(self, key, new_key):
        """Renames ``key`` to ``new_key`` if ``new_key`` does not yet exist.
        It returns an error under the same conditions as
        :meth:`~tredis.RedisClient.rename`.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to rename
        :type key: :class:`str`, :class:`bytes`
        :param new_key: The key to rename it to
        :type new_key: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'RENAMENX', key, new_key], 1)

    def restore(self, key, ttl, value, replace=False):
        """Create a key associated with a value that is obtained by
        deserializing the provided serialized value (obtained via
        :meth:`~tredis.RedisClient.dump`).

        If ``ttl`` is ``0`` the key is created without any expire, otherwise
        the specified expire time (in milliseconds) is set.

        :meth:`~tredis.RedisClient.restore` will return a
        ``Target key name is busy`` error when key already exists unless you
        use the :meth:`~tredis.RedisClient.restore` modifier (Redis 3.0 or
        greater).

        :meth:`~tredis.RedisClient.restore` checks the RDB version and data
        checksum. If they don't match an error is returned.

        .. note::

           **Time complexity**: ``O(1)`` to create the new key and additional
           ``O(N*M)`` to reconstruct the serialized value, where ``N`` is the
           number of Redis objects composing the value and ``M`` their average
           size. For small string values the time complexity is thus
           ``O(1)+O(1*M)`` where ``M`` is small, so simply ``O(1)``. However
           for sorted set values the complexity is ``O(N*M*log(N))`` because
           inserting values into sorted sets is ``O(log(N))``.

        :param key: The key to get the TTL for
        :type key: :class:`str`, :class:`bytes`
        :param int ttl: The number of seconds to set the timeout to
        :param value: The value to restore to the key
        :type value: :class:`str`, :class:`bytes`
        :param bool replace: Replace a pre-existing key
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        command = [b'RESTORE', key, ascii(ttl).encode('ascii'), value]
        if replace:
            command.append(b'REPLACE')
        return self._execute(command, b'OK')

    def scan(self, cursor=0, pattern=None, count=None):
        """The :meth:`~tredis.RedisClient.scan` command and the closely related
        commands :meth:`~tredis.RedisClient.sscan`,
        :meth:`~tredis.RedisClient.hscan` and :meth:`~tredis.RedisClient.zscan`
        are used in order to incrementally iterate over a collection of
        elements.

        - :meth:`~tredis.RedisClient.scan` iterates the set of keys in the
          currently selected Redis database.
        - :meth:`~tredis.RedisClient.sscan` iterates elements of Sets types.
        - :meth:`~tredis.RedisClient.hscan` iterates fields of Hash types and
          their associated values.
        - :meth:`~tredis.RedisClient.zscan` iterates elements of Sorted Set
          types and their associated scores.

        **Basic usage**

        :meth:`~tredis.RedisClient.scan` is a cursor based iterator.
        This means that at every call of the command, the server returns an
        updated cursor that the user needs to use as the cursor argument in
        the next call.

        An iteration starts when the cursor is set to ``0``, and terminates
        when the cursor returned by the server is ``0``.

        For more information on :meth:`~tredis.RedisClient.scan`,
        visit the `Redis docs on scan <http://redis.io/commands/scan>`_.

        .. note::

           **Time complexity**: ``O(1)`` for every call. ``O(N)`` for a
           complete iteration, including enough command calls for the cursor to
           return back to ``0``. ``N`` is the number of elements inside the
           collection.

        :param int cursor: The server specified cursor value or ``0``
        :param pattern: An optional pattern to apply for key matching
        :type pattern: :class:`str`, :class:`bytes`
        :param int count: An optional amount of work to perform in the scan
        :rtype: int, list
        :returns: A tuple containing the cursor and the list of keys
        :raises: :exc:`~tredis.exceptions.RedisError`

        """

        def format_response(value):
            """Format the response from redis

            :param tuple value: The return response from redis
            :rtype: tuple(int, list)

            """
            return int(value[0]), value[1]

        command = [b'SCAN', ascii(cursor).encode('ascii')]
        if pattern:
            command += [b'MATCH', pattern]
        if count:
            command += [b'COUNT', ascii(count).encode('ascii')]
        return self._execute(command, format_callback=format_response)

    def sort(self,
             key,
             by=None,
             external=None,
             offset=0,
             limit=None,
             order=None,
             alpha=False,
             store_as=None):
        """Returns or stores the elements contained in the list, set or sorted
        set at key. By default, sorting is numeric and elements are compared by
        their value interpreted as double precision floating point number.

        The ``external`` parameter is used to specify the
        `GET <http://redis.io/commands/sort#retrieving-external-keys>_`
        parameter for retrieving external keys. It can be a single string
        or a list of strings.

        .. note::

           **Time complexity**: ``O(N+M*log(M))`` where ``N`` is the number of
           elements in the list or set to sort, and ``M`` the number of
           returned elements. When the elements are not sorted, complexity is
           currently ``O(N)`` as there is a copy step that will be avoided in
           next releases.

        :param key: The key to get the refcount for
        :type key: :class:`str`, :class:`bytes`

        :param by: The optional pattern for external sorting keys
        :type by: :class:`str`, :class:`bytes`
        :param external: Pattern or list of patterns to return external keys
        :type external: :class:`str`, :class:`bytes`, list
        :param int offset: The starting offset when using limit
        :param int limit: The number of elements to return
        :param order: The sort order - one of ``ASC`` or ``DESC``
        :type order: :class:`str`, :class:`bytes`
        :param bool alpha: Sort the results lexicographically
        :param store_as: When specified, the key to store the results as
        :type store_as: :class:`str`, :class:`bytes`, None
        :rtype: list|int
        :raises: :exc:`~tredis.exceptions.RedisError`
        :raises: :exc:`ValueError`

        """
        if order and order not in [b'ASC', b'DESC', 'ASC', 'DESC']:
            raise ValueError('invalid sort order "{}"'.format(order))

        command = [b'SORT', key]
        if by:
            command += [b'BY', by]
        if external and isinstance(external, list):
            for entry in external:
                command += [b'GET', entry]
        elif external:
            command += [b'GET', external]
        if limit:
            command += [
                b'LIMIT',
                ascii(offset).encode('utf-8'),
                ascii(limit).encode('utf-8')
            ]
        if order:
            command.append(order)
        if alpha is True:
            command.append(b'ALPHA')
        if store_as:
            command += [b'STORE', store_as]

        return self._execute(command)

    def ttl(self, key):
        """Returns the remaining time to live of a key that has a timeout.
        This introspection capability allows a Redis client to check how many
        seconds a given key will continue to be part of the dataset.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to get the TTL for
        :type key: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'TTL', key])

    def type(self, key):
        """Returns the string representation of the type of the value stored at
        key. The different types that can be returned are: ``string``,
        ``list``, ``set``, ``zset``, and ``hash``.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key to get the type for
        :type key: :class:`str`, :class:`bytes`
        :rtype: bytes
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'TYPE', key])

    def wait(self, num_slaves, timeout=0):
        """his command blocks the current client until all the previous write
        commands are successfully transferred and acknowledged by at least the
        specified number of slaves. If the timeout, specified in milliseconds,
        is reached, the command returns even if the specified number of slaves
        were not yet reached.

        The command will always return the number of slaves that acknowledged
        the write commands sent before the :meth:`~tredis.RedisClient.wait`
        command, both in the case where the specified number of slaves are
        reached, or when the timeout is reached.

        .. note::

           **Time complexity**: ``O(1)``

        :param int num_slaves: Number of slaves to acknowledge previous writes
        :param int timeout: Timeout in milliseconds
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        command = [
            b'WAIT',
            ascii(num_slaves).encode('ascii'),
            ascii(timeout).encode('ascii')
        ]
        return self._execute(command)



"""Redis List Commands Mixin"""


class ListsMixin(object):
    """Redis List Commands Mixin"""

    def llen(self, key):
        """
        Returns the length of the list stored at key.

        :param key: The list's key
        :type key: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.TRedisException`

        If key does not exist, it is interpreted as an empty list and 0 is
        returned. An error is returned when the value stored at key is not a
        list.

        .. note::

           **Time complexity** ``O(1)``

        """
        return self._execute([b'LLEN', key])

    def lrange(self, key, start, end):
        """
        Returns the specified elements of the list stored at key.

        :param key: The list's key
        :type key: :class:`str`, :class:`bytes`
        :param int start: zero-based index to start retrieving elements from
        :param int end: zero-based index at which to stop retrieving elements

        :rtype: list
        :raises: :exc:`~tredis.exceptions.TRedisException`

        The offsets start and stop are zero-based indexes, with 0 being the
        first element of the list (the head of the list), 1 being the next
        element and so on.

        These offsets can also be negative numbers indicating offsets
        starting at the end of the list. For example, -1 is the last element
        of the list, -2 the penultimate, and so on.

        Note that if you have a list of numbers from 0 to 100,
        ``lrange(key, 0, 10)`` will return 11 elements, that is, the
        rightmost item is included. This may or may not be consistent with
        behavior of range-related functions in your programming language of
        choice (think Ruby's ``Range.new``, ``Array#slice`` or Python's
        :func:`range` function).

        Out of range indexes will not produce an error. If start is larger
        than the end of the list, an empty list is returned. If stop is
        larger than the actual end of the list, Redis will treat it like the
        last element of the list.

        .. note::

           **Time complexity** ``O(S+N)`` where ``S`` is the distance of
           start offset from ``HEAD`` for small lists, from nearest end
           (``HEAD`` or ``TAIL``) for large lists; and ``N`` is the number
           of elements in the specified range.

        """
        return self._execute([b'LRANGE', key, start, end])

    def ltrim(self, key, start, stop):
        """
        Crop a list to the specified range.

        :param key: The list's key
        :type key: :class:`str`, :class:`bytes`
        :param int start: zero-based index to first element to retain
        :param int stop: zero-based index of the last element to retain
        :returns: did the operation succeed?
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.TRedisException`

        Trim an existing list so that it will contain only the specified
        range of elements specified.

        Both `start` and `stop` are zero-based indexes, where 0 is the first
        element of the list (the head), 1 the next element and so on.
        For example: ``ltrim('foobar', 0, 2)`` will modify the list stored at
        ``foobar`` so that only the first three elements of the list will
        remain.

        `start` and `stop` can also be negative numbers indicating offsets
        from the end of the list, where -1 is the last element of the list,
        -2 the penultimate element and so on.

        Out of range indexes will not produce an error: if `start` is larger
        than the `end` of the list, or `start > end`, the result will be an
        empty list (which causes `key` to be removed). If `end` is larger
        than the end of the list, Redis will treat it like the last element
        of the list.

        A common use of LTRIM is together with LPUSH / RPUSH. For example::

            client.lpush('mylist', 'somelement')
            client.ltrim('mylist', 0, 99)

        This pair of commands will push a new element on the list, while
        making sure that the list will not grow larger than 100 elements.
        This is very useful when using Redis to store logs for example. It is
        important to note that when used in this way LTRIM is an O(1)
        operation because in the average case just one element is removed
        from the tail of the list.

        .. note::

           Time complexity: ``O(N)`` where `N` is the number of elements to
           be removed by the operation.

        """
        return self._execute([b'LTRIM', key, start, stop], b'OK')

    def lpush(self, key, *values):
        """
        Insert all the specified values at the head of the list stored at key.

        :param key: The list's key
        :type key: :class:`str`, :class:`bytes`
        :param values: One or more positional arguments to insert at the
            beginning of the list.  Each value is inserted at the beginning
            of the list individually (see discussion below).
        :returns: the length of the list after push operations
        :rtype: int
        :raises: :exc:`~tredis.exceptions.TRedisException`

        If `key` does not exist, it is created as empty list before
        performing the push operations. When key holds a value that is not a
        list, an error is returned.

        It is possible to push multiple elements using a single command call
        just specifying multiple arguments at the end of the command.
        Elements are inserted one after the other to the head of the list,
        from the leftmost element to the rightmost element. So for instance
        ``client.lpush('mylist', 'a', 'b', 'c')`` will result into a list
        containing ``c`` as first element, ``b`` as second element and ``a``
        as third element.

        .. note::

           **Time complexity**: ``O(1)``

        """
        return self._execute([b'LPUSH', key] + list(values))

    def lpushx(self, key, *values):
        """
        Insert values at the head of an existing list.

        :param key: The list's key
        :type key: :class:`str`, :class:`bytes`
        :param values: One or more positional arguments to insert at the
            beginning of the list.  Each value is inserted at the beginning
            of the list individually (see discussion below).
        :returns: the length of the list after push operations, zero if
            `key` does not refer to a list
        :rtype: int
        :raises: :exc:`~tredis.exceptions.TRedisException`

        This method inserts `values` at the head of the list stored at `key`,
        only if `key` already exists and holds a list. In contrary to
        :meth:`.lpush`, no operation will be performed when key does not yet
        exist.

        .. note::

           **Time complexity**: ``O(1)``

        """
        return self._execute([b'LPUSHX', key] + list(values))

    def lpop(self, key):
        """
        Removes and returns the first element of the list stored at key.

        :param key: The list's key
        :type key: :class:`str`, :class:`bytes`
        :returns: the element at the head of the list, :data:`None` if the
            list does not exist
        :raises: :exc:`~tredis.exceptions.TRedisException`

        .. note::

           **Time complexity**: ``O(1)``

        """
        return self._execute([b'LPOP', key])

    def rpush(self, key, *values):
        """
        Insert all the specified values at the tail of the list stored at key.

        :param key: The list's key
        :type key: :class:`str`, :class:`bytes`
        :param values: One or more positional arguments to insert at the
            tail of the list.
        :returns: the length of the list after push operations
        :rtype: int
        :raises: :exc:`~tredis.exceptions.TRedisException`

        If `key` does not exist, it is created as empty list before performing
        the push operation. When `key` holds a value that is not a list, an
        error is returned.

        It is possible to push multiple elements using a single command call
        just specifying multiple arguments at the end of the command.
        Elements are inserted one after the other to the tail of the list,
        from the leftmost element to the rightmost element. So for instance
        the command  ``client.rpush('mylist', 'a', 'b', 'c')`` will result
        in a list containing ``a`` as first element, ``b`` as second element
        and ``c`` as third element.

        .. note::

           **Time complexity**: ``O(1)``

        """
        return self._execute([b'RPUSH', key] + list(values))

    def rpushx(self, key, *values):
        """
        Insert values at the tail of an existing list.

        :param key: The list's key
        :type key: :class:`str`, :class:`bytes`
        :param values: One or more positional arguments to insert at the
            tail of the list.
        :returns: the length of the list after push operations or
            zero if `key` does not refer to a list
        :rtype: int
        :raises: :exc:`~tredis.exceptions.TRedisException`

        This method inserts value at the tail of the list stored at `key`,
        only if `key` already exists and holds a list. In contrary to
        method:`.rpush`, no operation will be performed when `key` does not
        yet exist.

        .. note::

           **Time complexity**: ``O(1)``

        """
        return self._execute([b'RPUSHX', key] + list(values))

    def rpop(self, key):
        """
        Removes and returns the last element of the list stored at key.

        :param key: The list's key
        :type key: :class:`str`, :class:`bytes`
        :returns: the length of the list after push operations or
            zero if `key` does not refer to a list
        :returns: the element at the tail of the list, :data:`None` if the
            list does not exist
        :rtype: int
        :raises: :exc:`~tredis.exceptions.TRedisException`

        """
        return self._execute([b'RPOP', key])



"""Redis Hash Commands Mixin"""
from tornado import concurrent


class HashesMixin(object):
    """Redis Hash Commands Mixin"""

    def hset(self, key, field, value):
        """Sets `field` in the hash stored at `key` to `value`.

        If `key` does not exist, a new key holding a hash is created. If
        `field` already exists in the hash, it is overwritten.

        .. note::

           **Time complexity**: always ``O(1)``

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :param field: The field in the hash to set
        :type key: :class:`str`, :class:`bytes`
        :param value: The value to set the field to
        :returns: ``1`` if `field` is a new field in the hash and `value`
            was set; otherwise, ``0`` if `field` already exists in the hash
            and the value was updated
        :rtype: int

        """
        return self._execute([b'HSET', key, field, value])

    def hget(self, key, field):
        """
        Returns the value associated with `field` in the hash stored at `key`.

        .. note::

           **Time complexity**: always ``O(1)``

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :param field: The field in the hash to get
        :type key: :class:`str`, :class:`bytes`
        :rtype: bytes, list
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'HGET', key, field])

    def hgetall(self, key):
        """
        Returns all fields and values of the has stored at `key`.

        The underlying redis `HGETALL`_ command returns an array of
        pairs.  This method converts that to a Python :class:`dict`.
        It will return an empty :class:`dict` when the key is not
        found.

        .. note::

           **Time complexity**: ``O(N)`` where ``N`` is the size
           of the hash.

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :returns: a :class:`dict` of key to value mappings for all
            fields in the hash

        .. _HGETALL: http://redis.io/commands/hgetall

        """

        def format_response(value):
            return dict(zip(value[::2], value[1::2]))

        return self._execute(
            [b'HGETALL', key], format_callback=format_response)

    def hmset(self, key, value_dict):
        """
        Sets fields to values as in `value_dict` in the hash stored at `key`.

        Sets the specified fields to their respective values in the hash
        stored at `key`.  This command overwrites any specified fields
        already existing in the hash.  If `key` does not exist, a new  key
        holding a hash is created.

        .. note::

           **Time complexity**: ``O(N)`` where ``N`` is the number of
           fields being set.

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :param value_dict: field to value mapping
        :type value_dict: :class:`dict`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        if not value_dict:
            future = concurrent.TracebackFuture()
            future.set_result(False)
        else:
            command = [b'HMSET', key]
            command.extend(sum(value_dict.items(), ()))
            future = self._execute(command)
        return future

    def hmget(self, key, *fields):
        """
        Returns the values associated with the specified `fields` in a hash.

        For every ``field`` that does not exist in the hash, :data:`None`
        is returned.  Because a non-existing keys are treated as empty
        hashes, calling :meth:`hmget` against a non-existing key will
        return a list of :data:`None` values.

        .. note::

           *Time complexity*: ``O(N)`` where ``N`` is the number of fields
           being requested.

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :param fields: iterable of field names to retrieve
        :returns: a :class:`dict` of field name to value mappings for
            each of the requested fields
        :rtype: dict

        """

        def format_response(val_array):
            return dict(zip(fields, val_array))

        command = [b'HMGET', key]
        command.extend(fields)
        return self._execute(command, format_callback=format_response)

    def hdel(self, key, *fields):
        """
        Remove the specified fields from the hash stored at `key`.

        Specified fields that do not exist within this hash are ignored.
        If `key` does not exist, it is treated as an empty hash and this
        command returns zero.

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :param fields: iterable of field names to retrieve
        :returns: the number of fields that were removed from the hash,
            not including specified by non-existing fields.
        :rtype: int

        """
        if not fields:
            future = concurrent.TracebackFuture()
            future.set_result(0)
        else:
            future = self._execute([b'HDEL', key] + list(fields))
        return future

    def hexists(self, key, field):
        """
        Returns if `field` is an existing field in the hash stored at `key`.

        .. note::

           *Time complexity*: ``O(1)``

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :param field: name of the field to test for
        :type key: :class:`str`, :class:`bytes`
        :rtype: bool

        """
        return self._execute([b'HEXISTS', key, field])

    def hincrby(self, key, field, increment):
        """
        Increments the number stored at `field` in the hash stored at `key`.

        If `key` does not exist, a new key holding a hash is created.  If
        `field` does not exist the value is set to ``0`` before the operation
        is performed.  The range of values supported is limited to 64-bit
        signed integers.

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :param field: name of the field to increment
        :type key: :class:`str`, :class:`bytes`
        :param increment: amount to increment by
        :type increment: int

        :returns: the value at `field` after the increment occurs
        :rtype: int

        """
        return self._execute(
            [b'HINCRBY', key, field, increment], format_callback=int)

    def hincrbyfloat(self, key, field, increment):
        """
        Increments the number stored at `field` in the hash stored at `key`.

        If the increment value is negative, the result is to have the hash
        field **decremented** instead of incremented.  If the field does not
        exist, it is set to ``0`` before performing the operation.  An error
        is returned if one of the following conditions occur:

        - the field contains a value of the wrong type (not a string)
        - the current field content or the specified increment are not
          parseable as a double precision floating point number

        .. note::

           *Time complexity*: ``O(1)``

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :param field: name of the field to increment
        :type key: :class:`str`, :class:`bytes`
        :param increment: amount to increment by
        :type increment: float

        :returns: the value at `field` after the increment occurs
        :rtype: float

        """
        return self._execute(
            [b'HINCRBYFLOAT', key, field, increment], format_callback=float)

    def hkeys(self, key):
        """
        Returns all field names in the hash stored at `key`.

        .. note::

           *Time complexity*: ``O(N)`` where ``N`` is the size of the hash

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :returns: the list of fields in the hash
        :rtype: list

        """
        return self._execute([b'HKEYS', key])

    def hlen(self, key):
        """
        Returns the number of fields contained in the hash stored at `key`.

        .. note::

           *Time complexity*: ``O(1)``

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :returns: the number of fields in the hash or zero when `key`
            does not exist
        :rtype: int

        """
        return self._execute([b'HLEN', key])

    def hsetnx(self, key, field, value):
        """
        Sets `field` in the hash stored at `key` only if it does not exist.

        Sets `field` in the hash stored at `key` only if `field` does not
        yet exist.  If `key` does not exist, a new key holding a hash is
        created.  If `field` already exists, this operation has no effect.

        .. note::

           *Time complexity*: ``O(1)``

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :param field: The field in the hash to set
        :type key: :class:`str`, :class:`bytes`
        :param value: The value to set the field to
        :returns: ``1`` if `field` is a new field in the hash and `value`
            was set.  ``0`` if `field` already exists in the hash and
            no operation was performed
        :rtype: int

        """
        return self._execute([b'HSETNX', key, field, value])

    def hvals(self, key):
        """
        Returns all values in the hash stored at `key`.

        .. note::

           *Time complexity* ``O(N)`` where ``N`` is the size of the hash

        :param key: The key of the hash
        :type key: :class:`str`, :class:`bytes`
        :returns: a :class:`list` of :class:`bytes` instances or an
            empty list when `key` does not exist
        :rtype: list

        """
        return self._execute([b'HVALS', key])



"""Redis String Commands Mixin"""

# Python 2 support for ascii()
if 'ascii' not in dir(__builtins__):  # pragma: nocover
    from tredis.compat import ascii

BITOP_AND = b'&'
"""Use for specifying a bitwise AND operation with
:meth:`~tredis.RedisClient.bitop`"""

BITOP_OR = b'|'
"""Use for specifying a bitwise OR operation with
:meth:`~tredis.RedisClient.bitop`"""

BITOP_XOR = b'^'
"""Use for specifying a bitwise XOR operation with
:meth:`~tredis.RedisClient.bitop`"""

BITOP_NOT = b'~'
"""Use for specifying a bitwise NOT operation with
:meth:`~tredis.RedisClient.bitop`"""

_BITOPTS = {
    BITOP_AND: b'AND',
    BITOP_OR: b'OR',
    BITOP_XOR: b'XOR',
    BITOP_NOT: b'NOT',
}


class StringsMixin(object):
    """Redis String Commands Mixin"""

    def append(self, key, value):
        """If key already exists and is a string, this command appends the
        value at the end of the string. If key does not exist it is created and
        set as an empty string, so :meth:`~tredis.RedisClient.append` will be
        similar to :meth:`~tredis.RedisClient.set` in this special case.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``. The amortized time complexity
           is ``O(1)`` assuming the appended value is small and the already
           present value is of any size, since the dynamic string library used
           by Redis will double the free space available on every reallocation.

        :param key: The key to get
        :type key: :class:`str`, :class:`bytes`
        :param value: The value to append to the key
        :type value: :class:`str`, :class:`bytes`
        :returns: The length of the string after the append operation
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'APPEND', key, value])

    def bitcount(self, key, start=None, end=None):
        """Count the number of set bits (population counting) in a string.

        By default all the bytes contained in the string are examined. It is
        possible to specify the counting operation only in an interval passing
        the additional arguments start and end.

        Like for the :meth:`~tredis.RedisClient.getrange` command start and
        end can contain negative values in order to index bytes starting from
        the end of the string, where ``-1`` is the last byte, ``-2`` is the
        penultimate, and so forth.

        Non-existent keys are treated as empty strings, so the command will
        return zero.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(N)``

        :param key: The key to get
        :type key: :class:`str`, :class:`bytes`
        :param int start: The start position to evaluate in the string
        :param int end: The end position to evaluate in the string
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`, :exc:`ValueError`

        """
        command = [b'BITCOUNT', key]
        if start is not None and end is None:
            raise ValueError('Can not specify start without an end')
        elif start is None and end is not None:
            raise ValueError('Can not specify start without an end')
        elif start is not None and end is not None:
            command += [ascii(start), ascii(end)]
        return self._execute(command)

    def bitop(self, operation, dest_key, *keys):
        """Perform a bitwise operation between multiple keys (containing
        string values) and store the result in the destination key.

        The values for operation can be one of:

            - ``b'AND'``
            - ``b'OR'``
            - ``b'XOR'``
            - ``b'NOT'``
            - :data:`tredis.BITOP_AND` or ``b'&'``
            - :data:`tredis.BITOP_OR` or ``b'|'``
            - :data:`tredis.BITOP_XOR` or ``b'^'``
            - :data:`tredis.BITOP_NOT` or ``b'~'``

        ``b'NOT'`` is special as it only takes an input key, because it
        performs inversion of bits so it only makes sense as an unary operator.

        The result of the operation is always stored at ``dest_key``.

        **Handling of strings with different lengths**

        When an operation is performed between strings having different
        lengths, all the strings shorter than the longest string in the set are
        treated as if they were zero-padded up to the length of the longest
        string.

        The same holds true for non-existent keys, that are considered as a
        stream of zero bytes up to the length of the longest string.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(N)``

        :param bytes operation: The operation to perform
        :param dest_key: The key to store the bitwise operation results to
        :type dest_key: :class:`str`, :class:`bytes`
        :param keys: One or more keys as keyword parameters for the bitwise op
        :type keys: :class:`str`, :class:`bytes`
        :return: The size of the string stored in the destination key, that is
                 equal to the size of the longest input string.
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`, :exc:`ValueError`

        """
        if (operation not in _BITOPTS.keys()
                and operation not in _BITOPTS.values()):
            raise ValueError('Invalid operation value: {}'.format(operation))
        elif operation in [b'~', b'NOT'] and len(keys) > 1:
            raise ValueError('NOT can only be used with 1 key')

        if operation in _BITOPTS.keys():
            operation = _BITOPTS[operation]

        return self._execute([b'BITOP', operation, dest_key] + list(keys))

    def bitpos(self, key, bit, start=None, end=None):
        """Return the position of the first bit set to ``1`` or ``0`` in a
        string.

        The position is returned, thinking of the string as an array of bits
        from left to right, where the first byte's most significant bit is at
        position 0, the second byte's most significant bit is at position
        ``8``, and so forth.

        The same bit position convention is followed by
        :meth:`~tredis.RedisClient.getbit` and
        :meth:`~tredis.RedisClient.setbit`.

        By default, all the bytes contained in the string are examined. It is
        possible to look for bits only in a specified interval passing the
        additional arguments start and end (it is possible to just pass start,
        the operation will assume that the end is the last byte of the string.
        However there are semantic differences as explained later). The range
        is interpreted as a range of bytes and not a range of bits, so
        ``start=0`` and ``end=2`` means to look at the first three bytes.

        Note that bit positions are returned always as absolute values starting
        from bit zero even when start and end are used to specify a range.

        Like for the :meth:`~tredis.RedisClient.getrange` command start and
        end can contain negative values in order to index bytes starting from
        the end of the string, where ``-1`` is the last byte, ``-2`` is the
        penultimate, and so forth.

        Non-existent keys are treated as empty strings.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(N)``

        :param key: The key to get
        :type key: :class:`str`, :class:`bytes`
        :param int bit: The bit value to search for (``1`` or ``0``)
        :param int start: The start position to evaluate in the string
        :param int end: The end position to evaluate in the string
        :returns: The position of the first bit set to ``1`` or ``0``
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`, :exc:`ValueError`

        """
        if 0 < bit > 1:
            raise ValueError('bit must be 1 or 0, not {}'.format(bit))
        command = [b'BITPOS', key, ascii(bit)]
        if start is not None and end is None:
            raise ValueError('Can not specify start without an end')
        elif start is None and end is not None:
            raise ValueError('Can not specify start without an end')
        elif start is not None and end is not None:
            command += [ascii(start), ascii(end)]
        return self._execute(command)

    def decr(self, key):
        """Decrements the number stored at key by one. If the key does not
        exist, it is set to 0 before performing the operation. An error is
        returned if the key contains a value of the wrong type or contains a
        string that can not be represented as integer. This operation is
        limited to 64 bit signed integers.

        See :meth:`~tredis.RedisClient.incr` for extra information on
        increment/decrement operations.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to decrement
        :type key: :class:`str`, :class:`bytes`
        :returns: The value of key after the decrement
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'DECR', key])

    def decrby(self, key, decrement):
        """Decrements the number stored at key by decrement. If the key does
        not exist, it is set to 0 before performing the operation. An error
        is returned if the key contains a value of the wrong type or contains
        a string that can not be represented as integer. This operation is
        limited to 64 bit signed integers.

        See :meth:`~tredis.RedisClient.incr` for extra information on
        increment/decrement operations.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to decrement
        :type key: :class:`str`, :class:`bytes`
        :param int decrement: The amount to decrement by
        :returns: The value of key after the decrement
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'DECRBY', key, ascii(decrement)])

    def get(self, key):
        """Get the value of key. If the key does not exist the special value
        :data:`None` is returned. An error is returned if the value stored
        at key is not a string, because :meth:`~tredis.RedisClient.get` only
        handles string values.

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to get
        :type key: :class:`str`, :class:`bytes`
        :rtype: bytes|None
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'GET', key])

    def getbit(self, key, offset):
        """Returns the bit value at offset in the string value stored at key.

        When offset is beyond the string length, the string is assumed to be a
        contiguous space with 0 bits. When key does not exist it is assumed to
        be an empty string, so offset is always out of range and the value is
        also assumed to be a contiguous space with 0 bits.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to get the bit from
        :type key: :class:`str`, :class:`bytes`
        :param int offset: The bit offset to fetch the bit from
        :rtype: bytes|None
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'GETBIT', key, ascii(offset)])

    def getrange(self, key, start, end):
        """Returns the bit value at offset in the string value stored at key.

        When offset is beyond the string length, the string is assumed to be a
        contiguous space with 0 bits. When key does not exist it is assumed to
        be an empty string, so offset is always out of range and the value is
        also assumed to be a contiguous space with 0 bits.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(N)`` where ``N`` is the length of
           the returned string. The complexity is ultimately determined by the
           returned length, but because creating a substring from an existing
           string is very cheap, it can be considered ``O(1)`` for small
           strings.

        :param key: The key to get the bit from
        :type key: :class:`str`, :class:`bytes`
        :param int start: The start position to evaluate in the string
        :param int end: The end position to evaluate in the string
        :rtype: bytes|None
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'GETRANGE', key, ascii(start), ascii(end)])

    def getset(self, key, value):
        """Atomically sets key to value and returns the old value stored at
        key. Returns an error when key exists but does not hold a string value.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to remove
        :type key: :class:`str`, :class:`bytes`
        :param value: The value to set
        :type value: :class:`str`, :class:`bytes`
        :returns: The previous value
        :rtype: bytes
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'GETSET', key, value])

    def incr(self, key):
        """Increments the number stored at key by one. If the key does not
        exist, it is set to ``0`` before performing the operation. An error is
        returned if the key contains a value of the wrong type or contains a
        string that can not be represented as integer. This operation is
        limited to 64 bit signed integers.

        .. note:: This is a string operation because Redis does not have a
           dedicated integer type. The string stored at the key is interpreted
           as a base-10 64 bit signed integer to execute the operation.

        Redis stores integers in their integer representation, so for string
        values that actually hold an integer, there is no overhead for storing
        the string representation of the integer.

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to increment
        :type key: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'INCR', key])

    def incrby(self, key, increment):
        """Increments the number stored at key by increment. If the key does
        not exist, it is set to 0 before performing the operation. An error is
        returned if the key contains a value of the wrong type or contains a
        string that can not be represented as integer. This operation is
        limited to 64 bit signed integers.

        See :meth:`~tredis.RedisClient.incr` for extra information on
        increment/decrement operations.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to increment
        :type key: :class:`str`, :class:`bytes`
        :param int increment: The amount to increment by
        :returns: The value of key after the increment
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'INCRBY', key, ascii(increment)])

    def incrbyfloat(self, key, increment):
        """Increment the string representing a floating point number stored at
        key by the specified increment. If the key does not exist, it is set to
        0 before performing the operation. An error is returned if one of the
        following conditions occur:

          - The key contains a value of the wrong type (not a string).
          - The current key content or the specified increment are not
            parsable as a double precision floating point number.

        If the command is successful the new incremented value is stored as the
        new value of the key (replacing the old one), and returned to the
        caller as a string.

        Both the value already contained in the string key and the increment
        argument can be optionally provided in exponential notation, however
        the value computed after the increment is stored consistently in the
        same format, that is, an integer number followed (if needed) by a dot,
        and a variable number of digits representing the decimal part of the
        number. Trailing zeroes are always removed.

        The precision of the output is fixed at 17 digits after the decimal
        point regardless of the actual internal precision of the computation.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to increment
        :type key: :class:`str`, :class:`bytes`
        :param float increment: The amount to increment by
        :returns: The value of key after the increment
        :rtype: bytes
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'INCRBYFLOAT', key, ascii(increment)])

    def mget(self, *keys):
        """Returns the values of all specified keys. For every key that does
        not hold a string value or does not exist, the special value nil is
        returned. Because of this, the operation never fails.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(N)`` where ``N`` is the number of
           keys to retrieve.

        :param keys: One or more keys as keyword arguments to the function
        :type keys: :class:`str`, :class:`bytes`
        :rtype: list
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'MGET'] + list(keys))

    def mset(self, mapping):
        """Sets the given keys to their respective values.
        :meth:`~tredis.RedisClient.mset` replaces existing values with new
        values, just as regular :meth:`~tredis.RedisClient.set`. See
        :meth:`~tredis.RedisClient.msetnx` if you don't want to overwrite
        existing values.

        :meth:`~tredis.RedisClient.mset` is atomic, so all given keys are set
        at once. It is not possible for clients to see that some of the keys
        were updated while others are unchanged.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(N)`` where ``N`` is the number of
           keys to set.

        :param dict mapping: A mapping of key/value pairs to set
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        command = [b'MSET']
        for key, value in mapping.items():
            command += [key, value]
        return self._execute(command, b'OK')

    def msetnx(self, mapping):
        """Sets the given keys to their respective values.
        :meth:`~tredis.RedisClient.msetnx` will not perform any operation at
        all even if just a single key already exists.

        Because of this semantic :meth:`~tredis.RedisClient.msetnx` can be used
        in order to set different keys representing different fields of an
        unique logic object in a way that ensures that either all the fields or
        none at all are set.

        :meth:`~tredis.RedisClient.msetnx` is atomic, so all given keys are set
        at once. It is not possible for clients to see that some of the keys
        were updated while others are unchanged.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(N)`` where ``N`` is the number of
           keys to set.

        :param dict mapping: A mapping of key/value pairs to set
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        command = [b'MSETNX']
        for key, value in mapping.items():
            command += [key, value]
        return self._execute(command, 1)

    def psetex(self, key, milliseconds, value):
        """:meth:`~tredis.RedisClient.psetex` works exactly like
        :meth:`~tredis.RedisClient.psetex` with the sole difference that the
        expire time is specified in milliseconds instead of seconds.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to set
        :type key: :class:`str`, :class:`bytes`
        :param int milliseconds: Number of milliseconds for TTL
        :param value: The value to set
        :type value: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute(
            [b'PSETEX', key, ascii(milliseconds), value], b'OK')

    def set(self, key, value, ex=None, px=None, nx=False, xx=False):
        """Set key to hold the string value. If key already holds a value, it
        is overwritten, regardless of its type. Any previous time to live
        associated with the key is discarded on successful
        :meth:`~tredis.RedisClient.set` operation.

        If the value is not one of :class:`str`, :class:`bytes`, or
        :class:`int`, a :exc:`ValueError` will be raised.

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to remove
        :type key: :class:`str`, :class:`bytes`
        :param value: The value to set
        :type value: :class:`str`, :class:`bytes`, :class:`int`
        :param int ex: Set the specified expire time, in seconds
        :param int px: Set the specified expire time, in milliseconds
        :param bool nx: Only set the key if it does not already exist
        :param bool xx: Only set the key if it already exist
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`
        :raises: :exc:`ValueError`

        """
        command = [b'SET', key, value]
        if ex:
            command += [b'EX', ascii(ex).encode('ascii')]
        if px:
            command += [b'PX', ascii(px).encode('ascii')]
        if nx:
            command.append(b'NX')
        if xx:
            command.append(b'XX')
        return self._execute(command, b'OK')

    def setbit(self, key, offset, bit):
        """Sets or clears the bit at offset in the string value stored at key.

        The bit is either set or cleared depending on value, which can be
        either 0 or 1. When key does not exist, a new string value is created.
        The string is grown to make sure it can hold a bit at offset. The
        offset argument is required to be greater than or equal to 0, and
        smaller than 2 :sup:`32` (this limits bitmaps to 512MB). When the
        string at key is grown, added bits are set to 0.

        .. warning:: When setting the last possible bit (offset equal to
           2 :sup:`32` -1) and the string value stored at key does not yet hold
           a string value, or holds a small string value, Redis needs to
           allocate all intermediate memory which can block the server for some
           time. On a 2010 MacBook Pro, setting bit number 2 :sup:`32` -1
           (512MB allocation) takes ~300ms, setting bit number 2 :sup:`30` -1
           (128MB allocation) takes ~80ms, setting bit number 2 :sup:`28` -1
           (32MB allocation) takes ~30ms and setting bit number 2 :sup:`26` -1
           (8MB allocation) takes ~8ms. Note that once this first allocation is
           done, subsequent calls to :meth:`~tredis.RedisClient.setbit` for the
           same key will not have the allocation overhead.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to get the bit from
        :type key: :class:`str`, :class:`bytes`
        :param int offset: The bit offset to fetch the bit from
        :param int bit: The value (``0`` or ``1``) to set for the bit
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        if 0 < bit > 1:
            raise ValueError('bit must be 1 or 0, not {}'.format(bit))
        return self._execute([b'SETBIT', key, ascii(offset), ascii(bit)])

    def setex(self, key, seconds, value):
        """Set key to hold the string value and set key to timeout after a
        given number of seconds.

        :meth:`~tredis.RedisClient.setex` is atomic, and can be reproduced by
        using :meth:`~tredis.RedisClient.set` and
        :meth:`~tredis.RedisClient.expire` inside an
        :meth:`~tredis.RedisClient.multi` /
        :meth:`~tredis.RedisClient.exec` block. It is provided as a faster
        alternative to the given sequence of operations, because this operation
        is very common when Redis is used as a cache.

        An error is returned when seconds is invalid.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to set
        :type key: :class:`str`, :class:`bytes`
        :param int seconds: Number of seconds for TTL
        :param value: The value to set
        :type value: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SETEX', key, ascii(seconds), value], b'OK')

    def setnx(self, key, value):
        """Set key to hold string value if key does not exist. In that case, it
        is equal to :meth:`~tredis.RedisClient.setnx`. When key already holds a
        value, no operation is performed. :meth:`~tredis.RedisClient.setnx` is
        short for "SET if Not eXists".

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to set
        :type key: :class:`str`, :class:`bytes`
        :param value: The value to set
        :type value: :class:`str`, :class:`bytes`, :class:`int`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SETNX', key, value], 1)

    def setrange(self, key, offset, value):
        """Overwrites part of the string stored at key, starting at the
        specified offset, for the entire length of value. If the offset is
        larger than the current length of the string at key, the string is
        padded with zero-bytes to make offset fit. Non-existing keys are
        considered as empty strings, so this command will make sure it holds a
        string large enough to be able to set value at offset.

        .. note:: The maximum offset that you can set is 2 :sup:`29` -1
           (536870911), as Redis Strings are limited to 512 megabytes. If you
           need to grow beyond this size, you can use multiple keys.

        .. warning:: When setting the last possible byte and the string value
           stored at key does not yet hold a string value, or holds a small
           string value, Redis needs to allocate all intermediate memory which
           can block the server for some time. On a 2010 MacBook Pro, setting
           byte number 536870911 (512MB allocation) takes ~300ms, setting byte
           number 134217728 (128MB allocation) takes ~80ms, setting bit number
           33554432 (32MB allocation) takes ~30ms and setting bit number
           8388608 (8MB allocation) takes ~8ms. Note that once this first
           allocation is done, subsequent calls to
           :meth:`~tredis.RedisClient.setrange` for the same key will not have
           the allocation overhead.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``, not counting the time taken to
           copy the new string in place. Usually, this string is very small so
           the amortized complexity is ``O(1)``. Otherwise, complexity is
           ``O(M)`` with ``M`` being the length of the value argument.

        :param key: The key to get the bit from
        :type key: :class:`str`, :class:`bytes`
        :param value: The value to set
        :type value: :class:`str`, :class:`bytes`, :class:`int`
        :returns: The length of the string after it was modified by the command
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SETRANGE', key, ascii(offset), value])

    def strlen(self, key):
        """Returns the length of the string value stored at key. An error is
        returned when key holds a non-string value

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)``

        :param key: The key to set
        :type key: :class:`str`, :class:`bytes`
        :returns: The length of the string at key, or 0 when key does not exist
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'STRLEN', key])



#!/usr/bin/env python
#

import setuptools

classifiers = ['Development Status :: 5 - Production/Stable',
               'Intended Audience :: Developers',
               'License :: OSI Approved :: BSD License',
               'Operating System :: OS Independent',
               'Programming Language :: Python :: 2',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3',
               'Programming Language :: Python :: 3.4',
               'Programming Language :: Python :: 3.5',
               'Programming Language :: Python :: 3.6',
               'Topic :: Communications', 'Topic :: Internet',
               'Topic :: Software Development :: Libraries',
               'Topic :: Database']

requirements = ['tornado>4.0,<5', 'hiredis>=0.2.0,<1']
tests_require = ['nose', 'mock', 'codecov', 'coverage']

setuptools.setup(name='tredis',
                 version='0.8.0',
                 description='An asynchronous Redis client for Tornado',
                 long_description=open('README.rst').read(),
                 author='Gavin M. Roy',
                 author_email='gavinmroy@gmail.com',
                 url='http://github.com/gmr/tredis',
                 packages=['tredis'],
                 package_data={'': ['LICENSE', 'README.rst']},
                 include_package_data=True,
                 install_requires=requirements,
                 tests_require=tests_require,
                 license='BSD',
                 classifiers=classifiers)
